//AEDII Universidad de Murcia, curso 2016-2017
//Ejemplo de divide y vencerás y de paralelismo OpenMP
//con un código de recorrido de cadenas
//El problema consiste en dada una cadena C de n enteros,
//encontrar la subcadena S de m elementos consecutivos de C
//en la que la suma de los enteros pares es máxima.
//Se devuelve la posición donde empieza la subcadena.
//En caso de varias subcadenas con el mismo valor máximo, 
//se devuelve la de más a la izquierda
//
//Se muestran cuatro implementaciones:
//-iterativa
//-paralela del algoritmo iterativo
//-divide y vencerás
//-paralela del método divide y vencerás
//
//Se explican en el código.
//Se recomienda seguir la explicación empezando por el programa principal
//y las funciones a las que se llama desde él.
//
//Compilar con
//g++ -O3 dv.cpp -fopenmp
// -O3 para optimización de código
// -fopenmp para usar paralelizmo OpenMP
//Si se quiere escribir los datos (para tamaños pequeños), incluir opción de compilación -DDEBUG
//
//Para practicas con paralelismo se puede hacer en vuestros propios equipos,
//con gcc a partir de la versión 4
//También en internet en https://www.tutorialspoint.com
//donde parece que usan hasta 24 núcleos.
//
//Y para practicar con otros ejemplos de paralelismo se puede pedir 
//una cuenta en calisto.inf.um.es para trabajar con los problemas OpenMP y MPI 
//del concurso de programación paralela (luna.inf.um.es) en un quadcore.
//Y hay un concurso basico que tiene ejemplos de mergesort.
  
#include <iostream>
#include <stdlib.h>

//Para la toma de tiempos
#include <sys/time.h>
#include <signal.h>

//Para poder usar paralelismo multihilo con OpenMP
//para incluir OpenMP hay que incluir en la compilación la opción -fopenmp (para gcc)
#include <omp.h>

using namespace std;

void generate(int a[],long long n)
{
//Genera entre 0 y 100. Se puede cambiar
  for(long long i=0;i<n;i++)
    a[i]=(int) (1.*rand()/RAND_MAX*100.5);
}

void escribir(int a[],long long n)
{
  for(long long i=0;i<n;i++)
    cout <<a[i]<<" ";
  cout<<endl;
}

long long iterativo(int *a,long long n,long long m)
{
  if(n<m)
    return -1;
  else
  {
    int maximo=0;
    long long pos_max=0;
    for(long long i=0;i<m;i++)
    {
      if((a[i]%2)==0)
        maximo+=a[i];
    }
    int valor=maximo;
    for(long long i=m;i<n;i++)
    {
      if((a[i-m]%2)==0)
        valor-=a[i-m];
      if((a[i]%2)==0)
        valor+=a[i];
      if(valor>maximo)
      {
        maximo=valor;
        pos_max=i-m+1;
      }
    }
    return pos_max;
  }
}

int sumar_pares(int a[],long long n)
{
  int suma=0;
  for(long long i=0;i<n;i++)
    if((a[i]%2)==0)
      suma+=a[i];
  return suma;
}

long long dv(int a[],long long n,long long m,long long base)
{
  long res1,res2,res;
//Esta es la función pequeño de dv
  if(n<=base)
//Se resuelve con el método no divide y vencerás
    return iterativo(a,n,m);
  else
  {
//Llamada a primer subproblema
    res1=dv(a,n/2,m,base);
//y a segundo subproblema. Se suma la posición en la que empieza el subproblema
    res2=dv(&a[n/2],n-n/2,m,base)+n/2;
//Combinación
//Se obtienen las sumas. Se podrían haber guardado y nos ahorraríamos volver a calcularlas
    int maximo=sumar_pares(&a[res1],m);
    int valor=sumar_pares(&a[res2],m);
    res=res1;
//Nos quedamos con el mayor de los dos
    if(valor>maximo)
    { 
      maximo=valor;
      res=res2;
    }
//En la frontera por la que se ha dividido el array se resuelve el problema
//con m posiciones más a la izquierda y a la derecha
    res2=iterativo(&a[n/2-m+1],2*m-2,m)+n/2-m+1;
    valor=sumar_pares(&a[res2],m);
//y nos quedamos con el mayor
    if(valor>maximo)
    { 
      maximo=valor;
      res=res2;
    }
    return res;
  }  
}

long long dv_paralelo(int a[],long long n,long long m,long long base,int nivel)
{
  long long res1,res2,res;
//Si hemos llegado al nivel máximo de paralelismo
  if(nivel<=1)
//Seguimos con divide y vencerás sin paralelismo
//También se podría usar el iterativo
    return dv(a,n,m,base);
  else
  {
//Se indica que en la siguiente región paralela se van a generar dos hilos,
//el que la pone en marcha y otro esclavo. Se llamarán 0 y 1.
//Los números de los hilos son relativos al grupo de hilos en el que están.
//Se repiten los números para grupos distintos.
    omp_set_num_threads(2);
#pragma omp parallel
    {
       int soy=omp_get_thread_num();
       if(soy==0)
//el primer hilo trabaja con la primera parte del array
         res1=dv_paralelo(a,n/2,m,base,nivel-1);
       else    
//y el otro con la parte final
         res2=dv_paralelo(&a[n/2],n-n/2,m,base,nivel-1)+n/2;
    } 
    int maximo=sumar_pares(&a[res1],m);
    int valor=sumar_pares(&a[res2],m);
    res=res1;
    if(valor>maximo)
    { 
      maximo=valor;
      res=res2;
    }
    res2=iterativo(&a[n/2-m+1],2*m-2,m)+n/2-m+1;
    valor=sumar_pares(&a[res2],m);
    if(valor>maximo)
    { 
      maximo=valor;
      res=res2;
    }
    return res;
  }  
}

long long iterativo_paralelo(int *a,long long n,long long m,int th)
{
  int soy; //Para el identificador de hilo. Se numeran 0,1,...numhilos-1
  long long tam=n/th; //Tamaño de la cadena que corresponde a cada hilo
  int resultado[th]; //Para las sumas óptimas en cada hilo
  long long posicion[th]; //Posiciones donde empiezan las cadenas óptimas en cada hilo

//Función OpenMP que establece el número de hilos que trabajan en la región paralela
  omp_set_num_threads(th);
//La variable soy es privada, pues cada hilo tendrá un identificador distinto
#pragma omp parallel private(soy)
//Este bloque es la región paralela. La ejecutan todos los hilos
  {
//Cada hilo ejecuta esto, y guarda su identificador
    soy=omp_get_thread_num();
    if(soy!=th-1)
//Si no soy el último hilo, se resuelve un problema empezando en la posicion tam*soy
//y con tamaño tam+m-1 para considerar las últimas subcadenas que solapan con la parte siguiente del array
//(no funcionará para todos los tamaños, pues si tam+m-1 sele del array dará error)
//al resultado se le suma la zona de comienzo del array con el que cada hilo trabaja
      posicion[soy]=iterativo(&a[tam*soy],tam+m-1,m)+tam*soy;
    else
//si soy el último hilo trabajo hasta el final del array
      posicion[soy]=iterativo(&a[tam*soy],n-tam*soy,m)+tam*soy;
  }
#ifdef DEBUG
//Para mostrar las soluciones que obtiene cada hilo
  cout <<"Posiciones: ";
  for(int i=0;i<th;i++)
    cout << posicion[i]<<" ";
  cout <<endl;
#endif

//Se ha acabado la región paralela.
//Ahora trabaja sólo un hilo
//obtiene el óptimo de todos los almacenados por todos los hilos
  for(int i=0;i<th;i++)
     resultado[i]=sumar_pares(&a[posicion[i]],m);
  int maximo=resultado[0];
  long long pos_max=posicion[0];
  for(int i=1;i<th;i++)
  {
     if(resultado[i]>maximo)
      {
        maximo=resultado[i];
        pos_max=posicion[i];
      }
  }
  return pos_max;
}

long long mseconds(){
  struct timeval t;
  gettimeofday(&t, NULL);
//Guarda en t los segundos y microsegundos transcurridos desde una fecha fija
//Se pasa todo a microsegundos
  return t.tv_sec*1000000 + t.tv_usec;
}

int main()
{
   long long n; //Número de enteros en la cadena
   long long m; //Tamaño de la subcadena 
   long long resultado; //Para el resultado que devuelven las funciones
   int nthreads; //Número de hilos máximo en las ejecuciones paralelas
   long long base; //Tamaño del caso base para divide y vencerás
   int niveles; //Niveles de recursión generando nuevos hilos en la versión paralela de divide y vencerás

   long long ti,tf; //Para tiempo inicial y final

//leer desde teclado o redirigir la entrada desde fichero
   cin >>n >>m>>nthreads>>base>>niveles;

   int *a=new int[n];
//Se generan datos aleatoriamente, entre 0 y 100 
   generate(a,n);

#ifdef DEBUG
//Esta es una zona de compilación condicional.
//Si se compila con la opción -D DEBUG
//se compila esta zona, y en otro caso no se incluye en la compilación.
//Cuando se está comprobando si funciona se escribe el array,
//para comprobar que la solución es correcta. Para tamaños pequeños.
  escribir(a,n);
#endif

//Toma de tiempos antes de empezar. Devuelve microsegundos
   ti=mseconds(); 
   resultado=iterativo(a,n,m);
   tf=mseconds();
//Toma de tiempo final, para restarlos y medir el tiempo empeado en la resolución del problema
        
//Escribimos para poder comparar las distintas soluciones y los tiempos
   cout << "Solucion iterativo, tamanos "<<n << " , "<<m<< " : "<< resultado<<endl;
   cout << "          tiempo: "<<tf-ti<<endl;

//Para la implementación paralela ejecutamos variando el número de hilos
//desde 1 hasta el máximo de hilos que se ha proporcionado.
//Así analizamos la explotación del paralelismo.
  for(int j=1;j<=nthreads;j++)
  {
    ti=mseconds(); 
    resultado=iterativo_paralelo(a,n,m,j);
    tf=mseconds();
        
     cout << "Solucion paralelo, tamanos "<<n << " , "<<m<< " , "<<j <<" threads: "<<resultado<<endl;
     cout << "          tiempo: "<<tf-ti<<endl;
  }

   ti=mseconds(); 
   resultado=dv(a,n,m,base);
   tf=mseconds();
        
   cout << "Solucion divide venceras, tamanos "<<n << " , "<<m<< " , base "<<base<<" : " <<resultado<<endl;
   cout << "          tiempo: "<<tf-ti<<endl;

//Con el divide y vencerás paralelo se hacen experimentos lanzando nuevos hilos hasta el nivel máximo que se haya indicado.
//Si se llega a nivel 4 se lanzan 2*2*2*2 hilos
  for(int j=1;j<=niveles;j++)
  {
    ti=mseconds(); 
//Para permitir paralelismo anidado: que unos hilos generen nuevos hilos
     omp_set_nested(1);
     resultado=dv_paralelo(a,n,m,base,j);
    tf=mseconds();
        
     cout << "Solucion divide y venceras paralelo, tamanos "<<n << " , "<<m<< " , base "<<base<<" , " << j <<" niveles de paralelismo: "<<resultado<<endl;
     cout << "          tiempo: "<<tf-ti<<endl;
  }

   delete[] a;
   return 0;
}

